#include "crazyflieworker.h"
#include <iostream>
#include <unistd.h>
#include <sys/stat.h>

CrazyflieWorker::CrazyflieWorker(QObject *parent) : QObject(parent), conn(NULL)
{
}

CrazyflieWorker::~CrazyflieWorker()
{
    disconnectBackend();
}

/**
 * @brief CrazyflieWorker::connectBackend
 * Command the worker to connect to the backend
 * will set the private class struct, conn, once done.
 * The connected() signal is emitted once complete.
 */
void CrazyflieWorker::connectBackend()
{
    if (conn == NULL) {
        conn = ucart_backendConnect();
        emit (connected());
    } else {
        qInfo() << "Attempted to connect crazyflieworker when already connected!";
    }
    isBackendConnected = true;
    loadParamIds();
    loadLogIds();
    QStringList paramGroupList;
    for(int i = 0; i < paramGroups.size(); i++) {
        QString qgName(paramGroups.at(i).groupName.c_str());
        paramGroupList.append(qgName);
    }
    emit(gotParamGroups(paramGroupList));
    QStringList logVariablesList;
    struct toc_group curGroup;
    for(int i = 0; i < logGroups.size(); i++) {
        curGroup = logGroups.at(i);
        for(int j = 0; j < logGroups.at(i).entryNames.size(); j++) {
            QString qvName(logGroups.at(i).groupName.c_str());
            qvName = qvName.append(".");
            qvName = qvName.append(logGroups.at(i).entryNames.at(j).c_str());
            logVariablesList.append(qvName);
        }
    }
    emit(gotLogVariables(logVariablesList));
}

/**
 * @brief CrazyflieWorker::disconnectBackend
 * Command the worker to disconnect from the backend,
 * sets the conn struct to null and emits the
 * disconnected() signal once complete
 */
void CrazyflieWorker::disconnectBackend()
{
    if (conn) {
         ucart_backendDisconnect(conn);
         conn = NULL;
         isBackendConnected = false;
         emit (disconnected());
    }
}

void CrazyflieWorker::setCurrAttSetpoint(float roll, float pitch, float yaw, float throttle)
{
    currAttitudeSetpoint.roll = roll;
    currAttitudeSetpoint.pitch = pitch;
    currAttitudeSetpoint.yaw = yaw;
    currAttitudeSetpoint.throttle = throttle;
}

void CrazyflieWorker::setCurrAttRateSetpoint(float rollRate, float pitchRate, float yawRate, float throttle)
{
    qInfo() << "set rate";
    currAttitudeRateSetpoint.roll = rollRate;
    currAttitudeRateSetpoint.pitch = pitchRate;
    currAttitudeRateSetpoint.yaw = yawRate;
    currAttitudeRateSetpoint.throttle = throttle;
}

void CrazyflieWorker::setCurrMixedSetpoint(float roll, float pitch, float yawRate, float throttle){
    qInfo() << "set rate";
    currMixedSetpoint.roll = roll;
    currMixedSetpoint.pitch = pitch;
    currMixedSetpoint.yaw = yawRate;
    currMixedSetpoint.throttle = throttle;
}



void CrazyflieWorker::sendAttRateSetpoint()
{

    if(isBackendConnected){
        qInfo() << " send att rate" << currAttitudeRateSetpoint.throttle << " " << currAttitudeRateSetpoint.roll;

        struct frontend_override_data data;
        data.enable = 2;    //attitude rate
        data.time = 0;
        data.throttle = currAttitudeRateSetpoint.throttle;
        data.roll = currAttitudeRateSetpoint.roll;
        data.pitch = currAttitudeRateSetpoint.pitch;
        data.yaw = currAttitudeRateSetpoint.yaw;

        //send setpoint
        if(frontend_setoutputoverride(conn, &data) != 0){
            qInfo() << "error sending att rate setpoint";
        }
    }
}


void CrazyflieWorker::sendAttSetpoint()
{

    if(isBackendConnected){
        qInfo() << " send att" << currAttitudeSetpoint.throttle;

        struct frontend_override_data data;
        data.enable = 1;    //attitude
        data.time = 0;
        data.throttle = currAttitudeSetpoint.throttle;
        data.roll = currAttitudeSetpoint.roll;
        data.pitch = currAttitudeSetpoint.pitch;
        data.yaw = currAttitudeSetpoint.yaw;

        //send setpoint
        if(frontend_setoutputoverride(conn, &data) != 0){
            qInfo() << "error sending att setpoint";
        }
    }
}

void CrazyflieWorker::sendMixedAttSetpoint(){


    if(isBackendConnected){

        struct frontend_override_data data;
        data.enable = 3;    //roll and pitch are attitude, yaw is attitude angle
        data.time = 0;
        data.throttle = currMixedSetpoint.throttle;
        data.roll = currMixedSetpoint.roll;
        data.pitch = currMixedSetpoint.pitch;
        data.yaw = currMixedSetpoint.yaw;
        
        qInfo() << "send mixed setpoint" << "roll:" << data.roll << " pitch:" << data.pitch;


        //send setpoint
        if(frontend_setoutputoverride(conn, &data) != 0){
            qInfo() << "error sending mixed att setpoint";
        }
    }
}

void CrazyflieWorker::getParamValue(QString paramName){
    struct toc_info info = paramNameLookup(paramName);

    struct frontend_param_data data;
    data.block = 0;
    data.param = info.id;
    data.value = info.value;

    if(frontend_getparam(conn, &data) == 0){
        emit(gotParamValue(paramName, data.value));
    }
}

void CrazyflieWorker::setParamValue(QString paramName, double value) {
    struct toc_info info = paramNameLookup(paramName);

    struct frontend_param_data data;
    data.block = 0;
    data.param = info.id;
    data.value = (float) value;
    
    if(frontend_setparam(conn, &data) != 0){
        //emit(gotParamValue(paramName, data.value));
        qInfo() << "setParamError";
    }
}

struct toc_info CrazyflieWorker::paramNameLookup(QString paramName){
    struct toc_info paramData = paramIds.value(paramName);
    return paramData;
}

std::string CrazyflieWorker::getLogFile(int type) {
    struct frontend_getlogfile_data logfile;
    char fname[256];
    logfile.command = type;
    logfile.name = fname;
    if(frontend_getlogfile(conn, &logfile)) {
        return "";
    }
    std::string result(logfile.name, strlen(logfile.name));
    return result;
}

void CrazyflieWorker::loadParamIds() {
    std::string filename = getLogFile(1);
    std::ifstream paramTOCFile;
    paramTOCFile.open(filename.c_str());
    std::string line;
    //header and top line
    std::getline(paramTOCFile, line);
    std::getline(paramTOCFile, line);
    //actual values;
    while(std::getline(paramTOCFile, line)) {
        struct toc_info values;
        std::stringstream ss(line);
        std::string temp_str;
        for(int i = 0; i < 4; i++) {
            std::getline(ss, temp_str, '\t');
            if(i == 0) {
                values.id = (int16_t) stoi(temp_str);
            }
            else if(i == 1) {
                values.dataType = (int8_t) stoi(temp_str);
            }
            else if(i == 2) {
                values.strGroup = temp_str;
            }
            else if(i == 3) {
                values.strName = temp_str;
            }
        }
        values.value = 0;

        int gFound = findParamGroup(values.strGroup);
        if(gFound == -1) {
            struct toc_group groupx;
            groupx.groupName = values.strGroup;
            groupx.entryNames.push_back(values.strName);
            paramGroups.push_back(groupx);
        }
        else {
            paramGroups.at(gFound).entryNames.push_back(values.strName);
        }

        std::string identifier = values.strGroup + "." + values.strName;
        QString Qidentifier(identifier.c_str());
        paramIds.insert(Qidentifier, values);
    }
    paramTOCFile.close();
}

void CrazyflieWorker::loadLogIds() {
    std::string filename = getLogFile(2);
    std::ifstream logTOCFile;
    logTOCFile.open(filename.c_str());
    std::string line;
    //top line
    std::getline(logTOCFile, line);
    std::getline(logTOCFile, line);
    //actual values;
    while(std::getline(logTOCFile, line)) {
        struct toc_info values;
        std::stringstream ss(line);
        std::string temp_str;
        for(int i = 0; i < 4; i++) {
            std::getline(ss, temp_str, '\t');
            if(i == 0) {
                values.id = (int16_t) stoi(temp_str);
            }
            else if(i == 1) {
                values.dataType = (int8_t) stoi(temp_str);
            }
            else if(i == 2) {
                values.strGroup = temp_str;
            }
            else if(i == 3) {
                values.strName = temp_str;
            }
        }
        values.value = 0;

        int gFound = findLogGroup(values.strGroup);
        if(gFound == -1) {
            struct toc_group groupx;
            groupx.groupName = values.strGroup;
            groupx.entryNames.push_back(values.strName);
            logGroups.push_back(groupx);
        }
        else {
            logGroups.at(gFound).entryNames.push_back(values.strName);;
        }

        std::string identifier = values.strGroup + "." + values.strName;
        QString Qidentifier(identifier.c_str());
        logIds.insert(Qidentifier, values);
    }
    logTOCFile.close();
}

int CrazyflieWorker::findParamGroup(std::string gName) {
    int result = -1;
    for(int i = 0; i < paramGroups.size(); i++) {
        if(!paramGroups.at(i).groupName.compare(gName)) {
            result = i;
            break;
        }
    }
    return result;
}

int CrazyflieWorker::findLogGroup(std::string gName) {
    int result = -1;
    for(int i = 0; i < logGroups.size(); i++) {
        if(!logGroups.at(i).groupName.compare(gName)) {
            result = i;
            break;
        }
    }
    return result;
}

void CrazyflieWorker::getGroupEntries(int box, QString gName) {
    std::string name = gName.toStdString();
    int index = findParamGroup(name);
    std::vector<std::string> entries = paramGroups.at(index).entryNames;
    QStringList entryList;
    for(int i = 0; i < entries.size(); i++) {
        QString cur(entries.at(i).c_str());
        entryList.append(cur);
    }
    emit(gotGroupEntries(box, entryList));
}
